// -----------------------------------------------------------------------------
//
// Copyright 2011(c) Analog Devices, Inc.
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//  - Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  - Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in
//    the documentation and/or other materials provided with the
//    distribution.
//  - Neither the name of Analog Devices, Inc. nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//  - The use of this software may or may not infringe the patent rights
//    of one or more patent holders.  This license does not release you
//    from the requirement that you obtain separate licenses from these
//    patent holders to use this software.
//  - Use of the software either in source or binary form, must be run
//    on or directly connected to an Analog Devices Inc. component.
//
// THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// INTELLECTUAL PROPERTY RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// -----------------------------------------------------------------------------
// FILE NAME :  AD7492.v
// MODULE NAME: AD7492
// AUTHOR :     acostina
// AUTHORS EMAIL : adrian.costina@analog.com
// -----------------------------------------------------------------------------
// SVN REVISION: 1346
// -----------------------------------------------------------------------------
// KEYWORDS : AD7492
// -----------------------------------------------------------------------------
// PURPOSE : Driver for the AD7492 1 MSPS, 4mW Internal REF and CLK,  12 Bit
//           Parallel ADC 
// -----------------------------------------------------------------------------
// REUSE ISSUES
// Reset Strategy      : Active low reset signal
// Clock Domains       : 1 clock, tested at 100MHz
// Critical Timing     : N/A
// Test Features       : N/A
// Asynchronous I/F    : N/A
// Instantiations      : N/A
// Synthesizable (y/n) : Y
// Target Device       : AD7492
// Other               : The driver is intended to be used for AD7492 ADCs
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

`timescale 1ns/100ps

//------------------------------------------------------------------------------
//----------- Module Declaration -----------------------------------------------
//------------------------------------------------------------------------------

module AD7492

//----------- Ports Declarations -----------------------------------------------
(
    //clock and reset signals
    input               fpga_clk_i,         //system clock
    input               reset_n_i,          //active low reset signal

    //IP control and data interface
    output reg  [15:0]  data_o,             //data read from the ADC
    output reg          data_rd_ready_o,    //when set to high the data read from the ADC is available on the data_o bus

    //ADC control and data interface
    input       [11:0]  adc_data_i,         //ADC data
    input               adc_busy_n_i,       //ADC busy signal
    output  reg         adc_rd_n_o,         //ADC RS signal
    output  reg         adc_cs_n_o,         //ADC CS signal
    output  reg         adc_convst_n_o      //ADC CONVST signal
);

//------------------------------------------------------------------------------
//----------- Registers Declarations -------------------------------------------
//------------------------------------------------------------------------------
reg [ 8:0]  adc_state;      //current state for the ADC control state machineAD7492_vlg_tst
reg [ 8:0]  adc_next_state; //next state for the ADC control state machine

reg [31:0]  adc_tcycle_cnt; //counts the number of FPGA clock cycles to determine when an ADC cycle is complete


//------------------------------------------------------------------------------
//----------- Local Parameters -------------------------------------------------
//------------------------------------------------------------------------------
//ADC states
parameter ADC_IDLE_STATE        = 9'b000000001; // default state
parameter ADC_WAKEUP_STATE      = 9'b000000010; // allows time for wakeup
parameter ADC_START_CNV         = 9'b000000100; // activates CONV signal
parameter ADC_END_CNV_STATE     = 9'b000001000; // deactivates CONV signal after busy is high
parameter ADC_CS_LOW_STATE      = 9'b000010000; // activates CS signal, after busy is low. Can be skipped if CS is always tied to low
parameter ADC_RD_LOW_STATE      = 9'b000100000; // activates RD signal
parameter ADC_READ_CNV_RESULT   = 9'b001000000; // reads data on the parallel port
parameter ADC_RD_HIGH_STATE     = 9'b010000000; // deactivates RD signal
parameter ADC_CS_HIGH_STATE     = 9'b100000000; // deactivates CS signal. Can be skipped if CS is always tied to low

//ADC timing
parameter real FPGA_CLOCK_FREQ  = 100;      //FPGA clock frequency [MHz]
parameter real ADC_CYCLE_TIME   = 1;        //time between two ADC conversions (Tcyc) [us]. 
parameter [31:0] ADC_CYCLE_CNT  = FPGA_CLOCK_FREQ * ADC_CYCLE_TIME - 1; 

//------------------------------------------------------------------------------
//----------- Assign/Always Blocks ---------------------------------------------
//------------------------------------------------------------------------------

//update the ADC timing counters
always @(posedge fpga_clk_i)
begin
    if(reset_n_i == 1'b0)
    begin
        adc_tcycle_cnt  <= 0;
    end
    else
    begin
        if(adc_state == ADC_IDLE_STATE)
        begin
            adc_tcycle_cnt <= ADC_CYCLE_CNT;
        end
        else if(adc_tcycle_cnt != 0)
        begin
            adc_tcycle_cnt <= adc_tcycle_cnt - 1;
        end
    end
end

//update the ADC current state and the control signals
always @(posedge fpga_clk_i)
begin
    if(reset_n_i == 1'b0)
    begin
        adc_state <= ADC_IDLE_STATE;
    end
    else
    begin
        adc_state   <= adc_next_state;
        data_o      <= data_o;
        case (adc_state)
            ADC_IDLE_STATE:
            begin
                adc_convst_n_o  <= 1'b1;
                adc_rd_n_o      <= 1'b1;
                adc_cs_n_o      <= 1'b1;
                data_rd_ready_o <= 1'b0;
            end
            ADC_WAKEUP_STATE:
            begin
                adc_convst_n_o  <= 1'b1;
                adc_rd_n_o      <= 1'b1;
                adc_cs_n_o      <= 1'b1;
                data_rd_ready_o <= 1'b0;
            end
            ADC_START_CNV:
            begin
                adc_convst_n_o  <= 1'b0;
                adc_rd_n_o      <= 1'b1;
                adc_cs_n_o      <= 1'b1;
                data_rd_ready_o <= 1'b0;
            end
            ADC_END_CNV_STATE:
            begin
                adc_convst_n_o  <= 1'b1;
                adc_rd_n_o      <= 1'b1;
                adc_cs_n_o      <= 1'b1;
                data_rd_ready_o <= 1'b0;
            end
            ADC_CS_LOW_STATE:
            begin
                adc_convst_n_o  <= 1'b1;
                adc_rd_n_o      <= 1'b0;
                adc_cs_n_o      <= 1'b0;
                data_rd_ready_o <= 1'b0;
            end
            ADC_RD_LOW_STATE:
            begin
                adc_convst_n_o  <= 1'b1;
                adc_rd_n_o      <= 1'b0;
                adc_cs_n_o      <= 1'b0;
                data_rd_ready_o <= 1'b0;
            end
            ADC_READ_CNV_RESULT:
            begin
                adc_convst_n_o  <= 1'b1;
                adc_rd_n_o      <= 1'b0;
                adc_cs_n_o      <= 1'b0;
                data_rd_ready_o <= 1'b0;
                data_o          <= {4'b0000,adc_data_i};
            end
            ADC_RD_HIGH_STATE:
            begin
                adc_convst_n_o  <= 1'b1;
                adc_rd_n_o      <= 1'b1;
                adc_cs_n_o      <= 1'b0;
                data_rd_ready_o <= 1'b1;
            end
            ADC_CS_HIGH_STATE:
            begin
                adc_convst_n_o  <= 1'b1;
                adc_rd_n_o      <= 1'b1;
                adc_cs_n_o      <= 1'b1;
                data_rd_ready_o <= 1'b0;
            end
        endcase
    end
end

//update the ADC next state
always @(adc_state, adc_tcycle_cnt, adc_busy_n_i)
begin
    adc_next_state <= adc_state;
    case (adc_state)
        ADC_IDLE_STATE:
        begin
            adc_next_state <= ADC_WAKEUP_STATE;
        end
        ADC_WAKEUP_STATE:
        begin
            adc_next_state <= ADC_START_CNV;
        end
        ADC_START_CNV:
        begin
            if ( adc_busy_n_i == 1'b1 )
            begin
                adc_next_state <= ADC_END_CNV_STATE;
            end
            else
            begin
                adc_next_state <= ADC_START_CNV;
            end
        end
        ADC_END_CNV_STATE:
        begin
            if( adc_busy_n_i == 1'b0 )
            begin
                adc_next_state <= ADC_CS_LOW_STATE;
            end
        end
        ADC_CS_LOW_STATE:
        begin
            adc_next_state <= ADC_RD_LOW_STATE;
        end
        ADC_RD_LOW_STATE:
        begin
            adc_next_state <= ADC_READ_CNV_RESULT;
        end
        ADC_READ_CNV_RESULT:
        begin
            adc_next_state <= ADC_RD_HIGH_STATE;
        end
        ADC_RD_HIGH_STATE:
        begin
            adc_next_state <= ADC_CS_HIGH_STATE;
        end
        ADC_CS_HIGH_STATE:
        begin
            if( adc_tcycle_cnt == 0 )
            begin
                adc_next_state <= ADC_IDLE_STATE;
            end
        end
        default:
        begin
            adc_next_state <= ADC_IDLE_STATE;
        end
    endcase
end
endmodule
