// -----------------------------------------------------------------------------
//
// Copyright 2011(c) Analog Devices, Inc.
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without modification, 
// are permitted provided that the following conditions are met:
//  - Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  - Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in
//    the documentation and/or other materials provided with the
//    distribution.
//  - Neither the name of Analog Devices, Inc. nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//  - The use of this software may or may not infringe the patent rights
//    of one or more patent holders.  This license does not release you
//    from the requirement that you obtain separate licenses from these
//    patent holders to use this software.
//  - Use of the software either in source or binary form, must be run
//    on or directly connected to an Analog Devices Inc. component.
//
// THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED 
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY 
// AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
// INTELLECTUAL PROPERTY RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// -----------------------------------------------------------------------------
// FILE NAME :      ad7091top.v
// MODULE NAME :    ad7091top
// AUTHOR :         acozma
// AUTHORS EMAIL : andrei.cozma@analog.com
// -----------------------------------------------------------------------------
// SVN REVISION: $WCREV$
// -----------------------------------------------------------------------------
// KEYWORDS : ad7091
// -----------------------------------------------------------------------------
// PURPOSE : This module is intended to read data from ADCs, connected to a CED1Z 
//           board
// -----------------------------------------------------------------------------
// REUSE ISSUES
// Reset Strategy      : Active high reset signal
// Clock Domains       : 2 clocks - the system clock that drives the internal logic 
//                     : and a clock for ADC conversions
// Critical Timing     : N/A
// Test Features       : N/A
// Asynchronous I/F    : N/A
// Instantiations      : AD7091, write_master
// Synthesizable (y/n) : Y
// Target Device       : CED1Z
// Other               : Avalon accesible registers:
//  CONTROL_REGISTER    : address 0:    bit 0 - start acquisition
//                                      bit 1 - software reset
//                                      bit 2 - control_fixed_location (for 1, the avalon master will write data to a single location)
//                                      bit 3 - write data to the DUT
//  ACQ_COUNT_REGISTER  : address 1: counter to store the total number of sample acquisitions requested by software
//  BASE_REGISTER       : address 2: the base address where data will be written
//  STATUS_REGISTER     : address 3: status register
//  DUT_WRITE_REGISTER  : address 4: data to be written to the DUT 
// -----------------------------------------------------------------------------

`timescale 1ns/100ps //Use a timescale that is best for simulation.

//------------------------------------------------------------------------------
//----------- Module Declaration -----------------------------------------------
//------------------------------------------------------------------------------

module ad7091top
#(
    //data path width
    //range : 8, 16, 32, 64, 128, 256, 512, 1024 
    parameter DATAWIDTH = 32,

    // (DATAWIDTH)/8
    //range: 1, 2, 4, 8, 16, 32, 64, 128
    parameter BYTEENABLEWIDTH = 4,

    //The number of address bits exposed to the system interconnect fabric.
    //range: 1-32
    parameter ADDRESSWIDTH = 32,

    //FIFO depth of the internal buffer
    //range: 4 ,8, 16, 62, 64, 128
    parameter FIFODEPTH = 32,

    //Log2(FIFODEPTH)
    //range: 2-7
    parameter FIFODEPTH_LOG2 = 5,

    //Set to '1' to use on chip memory for the internal buffer
    //Set to '0' to use logical elements instead of memory
    //range: 0/1
    parameter FIFOUSEMEMORY = 1
)

//----------- Ports Declarations -----------------------------------------------
(
    //clock and reset signals
    input                           clk_i,
    input                           reset_i,
    input                           adc_clk_i,

     //avalon ports slave
    input       [31:0]              avalon_writedata_i,
    input                           avalon_write_i,
    input                           avalon_read_i,
    input       [ 2:0]              avalon_address_i,
    output  reg [31:0]              avalon_readdata_o,

    //avalon master ports
    input                           avalon_master_waitrequest_i,
    output  [DATAWIDTH-1:0]         avalon_master_address_o,
    output                          avalon_master_write_o,
    output  [BYTEENABLEWIDTH-1:0]   avalon_master_byteenable_o,
    output  [DATAWIDTH-1:0]         avalon_master_writedata_o,

    //DUT control and data interface
    input               adc_sdata_i,
    output              adc_sclk_o,
    output              adc_cs_n_o,
    output              adc_convst_n_o
);

//------------------------------------------------------------------------------
//----------- Registers Declarations -------------------------------------------
//------------------------------------------------------------------------------

// Avalon registers
reg [31:0] control_register;        //address 0: //bit 0 - start acquisition,
                                                 //bit 1 - software reset,
                                                 //bit 2 - control_fixed_location (for 1, the avalon master will write data to a single location)
                                                 //bit 3 - write data to the DUT
reg [31:0] acq_count_register;      //address 1: counter to store the total number of sample acquisitions requested by software
reg [31:0] base_register;           //address 2: the base address where data will be written
reg [31:0] status_register;         //address 3: status register
reg [31:0] dut_write_register;      //address 4: data to be written to the DUT

// Control inputs and outputs for write master
reg        control_go;             //strobe to initiate a write master transaction
reg        control_go_d1;          //delayed signal to be used to update the base for the next write request

// User logic inputs and outputs for write master
reg                 user_write_buffer;  //signal to activate the WRReq signal on the memory subcomponent of the write master core
reg [DATAWIDTH-1:0] user_buffer_data;   //data to be written in the memory subcomponent of the write master core

//Internal signals
reg [31:0]  length;                 //variable used to indicate the number of words to be writen by the Write master
reg [31:0]  committed_data;         //integer used to keep track of the number of words already written in the system memory
reg [23:0]  dev_total_read_cnt;     //counts the total number of samples received from the device
reg         run_acq;                //controls the start/stop of the data acquisition process
reg         acquisition_complete;   //signals that the acquisition is complete
reg         flag;                   //used for concatenation of two samples on a 32 bit memory word
reg         flag_d1;                //one clock delayed flag
reg         buffer_overflow;        //used to indicate that the buffer was overflown, and data is not contiguous
reg         write_addr_err;         //used to signal that a write request on a wrong address has been performed

//IP control and data interface
reg          dut_wr_data_n;
reg [15:0]   dut_data_i;
reg [15:0]   dut_data_o_d1;
reg          dut_data_rd_ready_d1;
reg          dut_data_wr_ready_d1;

//------------------------------------------------------------------------------
//----------- Wire Declarations ------------------------------------------------
//------------------------------------------------------------------------------

wire reset;                                     //used to concatenate the software/hardware reset signals in a single signal

//write master wires
wire                    control_fixed_location; //this only makes sense to enable when MAXBURSTCOUNT = 1
wire [ADDRESSWIDTH-1:0] control_write_base;     //address of the start location where data will be written
wire [ADDRESSWIDTH-1:0] control_write_length;   //number of bytes to be written by the write master 
wire                    control_done;
wire                    user_buffer_full;       //signal from the write master core, active in case the memory subcomponent is full

//IP control and data interfaces
wire          dut_data_rd_ready;
wire          dut_data_wr_ready;
wire [15:0]   dut_data_o;

//------------------------------------------------------------------------------
//----------- Assign/Always Blocks ---------------------------------------------
//------------------------------------------------------------------------------

assign control_fixed_location   = control_register[2];              //1 if data will be written to the same address
assign control_write_base       = base_register + committed_data;   //the base address sent to the memory write master will be increased after each successful write
assign control_write_length     = length;                           //for each write, a variable number of data will be written, as a function of acquisition speed and avalaon transfer speed
assign reset                    = reset_i | control_register[1];    //merge software reset and hardware reset in the same signal

//process to update the avalon registers
always @(posedge clk_i)
begin
    control_register    <= 32'h00;
    status_register[0]  <= acquisition_complete;
    status_register[1]  <= buffer_overflow;
    status_register[2]  <= write_addr_err;
    if (reset == 1'b1)
    begin
        control_register    <= 32'b0;
        acq_count_register  <= 32'b0;
        base_register       <= 32'b0;
        status_register     <= 32'b0;
        dut_write_register  <= 32'b0;
        write_addr_err      <= 1'b0;
    end
    else if (avalon_write_i == 1'b1)
    begin
        case (avalon_address_i)
            3'b000:
            begin
                control_register    <= avalon_writedata_i;
            end
            3'b001: 
            begin
                acq_count_register  <= avalon_writedata_i;
            end
            3'b010:
            begin
                base_register       <= avalon_writedata_i;
            end
            3'b100:
           begin
                dut_write_register  <= avalon_writedata_i;
            end
            default:
            begin
                write_addr_err     <= 1'b1;
            end
        endcase
    end
    else if (avalon_read_i == 1'b1)
    begin
         case (avalon_address_i)
            3'b000:
            begin
                avalon_readdata_o   <= control_register;
            end
            3'b001:
            begin
                avalon_readdata_o   <= acq_count_register;
            end
            3'b010:
            begin
                avalon_readdata_o   <= base_register;
            end
            3'b011:
            begin
                avalon_readdata_o   <= status_register;
            end
            3'b100:
            begin
                avalon_readdata_o   <= dut_write_register;
            end
            default:
            begin
                avalon_readdata_o   <= status_register;
            end
        endcase
    end
end

//process to write data to the DUT
always @(posedge clk_i)
begin
    if(reset == 1'b1)
    begin
        dut_wr_data_n        <= 1'b1;
        dut_data_wr_ready_d1 <= 1'b1;
        dut_data_i           <= 16'b0;
    end
    else 
    begin
        dut_data_wr_ready_d1 <= dut_data_wr_ready;
        dut_data_i           <= dut_write_register[15:0];
        if(control_register[3] == 1)
        begin
            dut_wr_data_n <= 1'b0;
        end
        else if(dut_wr_data_n == 1'b0 && dut_data_wr_ready_d1 == 1'b0 && dut_data_wr_ready == 1'b1)
        begin
           dut_wr_data_n <= 1'b1; 
        end
    end
end

//process to read data from the DUT
always @(negedge clk_i)
begin
    if (reset == 1'b1)
    begin
        user_write_buffer   <= 1'b0;
        user_buffer_data    <= 32'h0;
        dev_total_read_cnt  <= 0;
        buffer_overflow     <= 0;
        flag                <= 1'b0;
        flag_d1             <= 1'b0;
    end 
    else if (run_acq == 1 )
    begin
        dut_data_rd_ready_d1 <= dut_data_rd_ready;
        flag_d1              <= flag;
        if (user_buffer_full == 1'b0 )
        begin
            if (dut_data_rd_ready_d1 == 1'b0 && dut_data_rd_ready == 1'b1)
            begin
                flag <= ~flag;
                if (flag == 1)
                begin
                    dev_total_read_cnt  <= dev_total_read_cnt + 24'b1;
                    user_write_buffer   <= 1'b1 ;
                    user_buffer_data    <= {dut_data_o[15:0],dut_data_o_d1[15:0]};
                end
                else
                begin
                    dev_total_read_cnt  <= dev_total_read_cnt + 24'b1;
                    user_write_buffer   <= 1'b0;
                    dut_data_o_d1       <= dut_data_o;
                end
            end
            else
            begin
                user_write_buffer   <= 1'b0 ;
                user_buffer_data    <= 32'hfefefefe;
            end
        end
        else
        begin
            buffer_overflow     <= 1;
            user_write_buffer   <= 1'b0 ;
            user_buffer_data    <= 32'hfefefefe;
        end
    end
    else
    begin
        user_write_buffer   <= 1'b0;
        user_buffer_data    <= 32'hfefefefe;
    end
end

//process to control the length and committed data signals
always @(posedge clk_i)
begin
    if (reset === 1'b1)
    begin
        committed_data <= 0;
        length         <= 0;
    end
    else if (control_go_d1 == 1'b1)
    begin
        committed_data  <= committed_data + length ;
        length          <= (flag_d1 == 1'b0 && flag == 1'b1) ? 32'h1 : 32'h0;
    end
    else if(flag_d1 == 1'b0 && flag == 1'b1)
    begin
        length  <= length + 32'h1;
    end
end

//process used to control the avalon master control_go signal
always @(posedge clk_i)
begin
    if (reset == 1'b1)
    begin
        control_go      <= 1'b0;
        control_go_d1   <= 1'b0;
    end
    else if (length > 0 && control_done == 1'b1)
    begin
        control_go      <= 1'b1;
        control_go_d1   <= control_go;
    end
    else
    begin
        control_go      <= 1'b0;
        control_go_d1   <= control_go;
    end
end

//process to control the run_acq signal
always @(posedge clk_i)
begin
    if ( reset == 1'b1 )
    begin
        run_acq                 <= 1'b0;
        acquisition_complete    <= 1'b0;
    end
    else if (control_register [0] == 1'b1)
    begin
        run_acq                 <= 1'b1;
        acquisition_complete    <= 1'b0;
    end
    else if (acq_count_register == dev_total_read_cnt)
    begin
        run_acq                 <= 1'b0;
        acquisition_complete    <= 1'b1;
    end
end

//------------------------------------------------------------------------------
//----------- Modules Instantiations -------------------------------------------
//------------------------------------------------------------------------------
write_master #(
    .DATAWIDTH(DATAWIDTH),
    .BYTEENABLEWIDTH(BYTEENABLEWIDTH),
    .ADDRESSWIDTH(ADDRESSWIDTH),
    .FIFODEPTH(FIFODEPTH),
    .FIFODEPTH_LOG2(FIFODEPTH_LOG2),
    .FIFOUSEMEMORY(FIFOUSEMEMORY) 
    ) w_m_i1(
    .clk(clk_i),
    .reset (reset),
    .control_fixed_location(control_fixed_location),
    .control_write_base(control_write_base),
    .control_write_length (control_write_length),
    .control_go(control_go),
    .control_done(control_done),
    .user_write_buffer(user_write_buffer),
    .user_buffer_data(user_buffer_data),
    .user_buffer_full(user_buffer_full),
    .master_address(avalon_master_address_o),
    .master_write(avalon_master_write_o),
    .master_byteenable(avalon_master_byteenable_o),
    .master_writedata(avalon_master_writedata_o),
    .master_waitrequest(avalon_master_waitrequest_i)
    );

    AD7091 ad7091_i1(
    //clock and reset signals
    .fpga_clk_i(clk_i),
    .adc_clk_i(adc_clk_i),
    .reset_n_i(~reset),

    //IP control and data interface
    .data_o(dut_data_o),
    .data_rd_ready_o(dut_data_rd_ready),

    //AD7980 control and data interface
    .adc_sdata_i(adc_sdata_i),
    .adc_sclk_o(adc_sclk_o),
    .adc_cs_n_o(adc_cs_n_o),
    .adc_convst_n_o(adc_convst_n_o)    
    );

endmodule

